/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package nyanner;

import com.almworks.sqlite4java.SQLiteConnection;
import com.almworks.sqlite4java.SQLiteException;
import com.almworks.sqlite4java.SQLiteStatement;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.Scanner;
import java.util.logging.Level;
import java.util.logging.Logger;
import javafx.application.Platform;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.geometry.Orientation;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.control.Alert;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Dialog;
import javafx.scene.control.Label;
import javafx.scene.control.ScrollBar;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableRow;
import javafx.scene.control.TableView;
import javafx.scene.control.TextField;
import javafx.scene.control.Tooltip;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.image.Image;
import javafx.scene.input.Clipboard;
import javafx.scene.input.ClipboardContent;
import javafx.scene.input.KeyCode;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.GridPane;
import javafx.stage.FileChooser;
import org.sqlorm.querybuilder.SelectBuilder;
import uriSchemeHandler.CouldNotOpenUriSchemeHandler;
import uriSchemeHandler.URISchemeHandler;

/**
 *
 * @author Yanju
 */
public class FXMLDocumentController implements Initializable {

    private Properties prop;
    private File dbPath;
    private File nyaaPath;
    private SQLiteConnection db;
    private ObservableList<Torrent> torrentList;
    private HashMap<String, Image> categoryImages;
    private HashMap<String, String> magnets;
    private HashMap<String, String> categoryCodes;
    private Image magnetIcon;
    private Dialog dbChooserDialog;
    private Alert help;
    private final FileChooser fileChooser = new FileChooser();
    private TextField viewField;
    private TextField nyaaField;
    private final Tooltip copyTooltip = new Tooltip("Magnet copied to clipboard");
    private SQLiteStatement st;
    private ScrollBar vertical;

    @FXML
    private ComboBox<?> categoriesField;

    @FXML
    private TextField searchField;

    @FXML
    private TableView<Torrent> itemsTable;

    @FXML
    private TableColumn<Torrent, Image> categoryColumn;

    @FXML
    private TableColumn<Torrent, String> nameColumn;

    @FXML
    private TableColumn<Torrent, String> dlColumn;

    @FXML
    private TableColumn<Torrent, String> sizeColumn;

    /**
     * Handles SQLiteStatement generation Torrent item generation and scrolling
     *
     * TODO - decoupling
     */
    public void search() {
        //fetches itemsTable ScrollBar adds scroll listener for infinite scrolling
        if (vertical == null) {
            vertical = getVerticalScrollbar(itemsTable);
            vertical.valueProperty().addListener((obs, oldV, newV) -> {
                // if scrollbar value over ~66% load 5 items while there are more items
                if (newV.doubleValue() > (vertical.getMax() / 1.3) && st != null) {

                    try {
                        if (!st.hasRow()) {

                            st.dispose();
                        } else {
                            int i = 0;
                            double targetValue = newV.doubleValue() * itemsTable.getItems().size();
                            while (st.step() && i < 5) {
                                torrentList.add(new Torrent(st.columnString(0), st.columnString(1), st.columnString(2), categoryImages.get(st.columnString(4)), st.columnString(15)));
                                i++;
                            }
                            vertical.setValue(targetValue / itemsTable.getItems().size());
                        }

                    } catch (SQLiteException ex) {
                        Logger.getLogger(FXMLDocumentController.class.getName()).log(Level.SEVERE, null, ex);
                    }
                }
            });

        }
        //resets scrollbar up
        vertical.setValue(vertical.getMin());

        //SQL query generation via SQLOrm-1.0.11 library
        String searchText = searchField.getText();
        String[] splited = searchText.split("\\s+");
        SelectBuilder qry = new SelectBuilder();
        for (String s : splited) {
            qry.select("*").from("nyaa").where().and("nyaa.name like '%" + s + "%'");
        }
        String category = categoriesField.getSelectionModel().getSelectedItem().toString();
        if (!category.equals("All categories")) {
            qry.select("*").from("nyaa").where().and(categoryCodes.get(category));
        }
        //SQLiteStatement creation and initial itemset fetching
        try {
            if (st != null) {
                st.dispose();
            }
            st = db.prepare(qry.toSql());
            torrentList.clear();
            int i = 0;
            while (st.step() && i < 20) {
                torrentList.add(new Torrent(st.columnString(0), st.columnString(1), st.columnString(2), categoryImages.get(st.columnString(4)), st.columnString(15)));
                i++;
            }
            if (!st.hasRow()) {
                st.dispose();
            }

        } catch (SQLiteException ex) {
            Logger.getLogger(FXMLDocumentController.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void disposeDB() {
        db.dispose();
    }

    /**
     * magnet URI handling via URISChemeHandler-1.7 library
     *
     * @param id id of the magnet to generate
     */
    public void download(String id) {

        try {
            String os = System.getProperty("os.name").toLowerCase();
            Runtime rt = Runtime.getRuntime();
            String url = ("magnet:?xt=urn:btih:" + magnets.get(id));
            //fallback copy to clipboard in case uri handler fails
            Clipboard clpbrd = Clipboard.getSystemClipboard();
            final ClipboardContent content = new ClipboardContent();
            content.putString(url);
            clpbrd.setContent(content);

            //--
            URI magnetLinkUri = new URI(url);
            URISchemeHandler uriSchemeHandler = new URISchemeHandler();
            uriSchemeHandler.open(magnetLinkUri);
        } catch (URISyntaxException | CouldNotOpenUriSchemeHandler ex) {
            Logger.getLogger(FXMLDocumentController.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * closes DB and exits
     */
    public void exit() {
        db.dispose();
        Platform.exit();
        System.exit(0);

    }

    /**
     * DB files chooser dialog
     *
     * TODO - decouple save and set
     */
    public void showDbChooserDialog() {
        dbChooserDialog.showAndWait().filter(reply -> reply == ButtonType.OK)
                .ifPresent(respuesta -> {
                    if (!viewField.getText().equals(dbPath.toString())) {
                        dbPath = new File(viewField.getText());
                    }
                    if (!nyaaField.getText().equals(nyaaPath.toString())) {
                        nyaaPath = new File(nyaaField.getText());
                    }

                });
        if (nyaaPath == null || dbPath == null) {
            exit();
        }
        saveProperties();
        setBD();
    }

    /**
     * saves the DB files location in the config file for next time
     */
    private void saveProperties() {
        OutputStream output = null;
        try {

            output = new FileOutputStream("config.properties");

            // set the properties value
            prop.setProperty("dbPath", dbPath.toString());
            prop.setProperty("nyaaPath", nyaaPath.toString());
            // save properties to project root folder
            prop.store(output, null);

        } catch (IOException io) {
        } finally {
            if (output != null) {
                try {
                    output.close();
                } catch (IOException e) {
                }
            }
        }
    }

    /**
     * shows help/about pane
     */
    public void showHelp() {
        help.showAndWait();
    }

    /**
     * Initializes and opens connection to the choosen DB file Maps magnet-id
     * file for quick fetching
     */
    private void setBD() {
        try {
            db = new SQLiteConnection(dbPath);
            db.open(true);
            magnets = new HashMap<>();
            FileInputStream inputStream = null;
            try {
                inputStream = new FileInputStream(nyaaPath);
                Scanner inFile = new Scanner(inputStream, "UTF-8");
                String token;
                String[] splitToken;
                while (inFile.hasNext()) {
                    token = inFile.next();
                    splitToken = token.split("\\|");
                    magnets.put(splitToken[0], splitToken[2]);
                }
            } catch (FileNotFoundException ex) {
                Logger.getLogger(FXMLDocumentController.class.getName()).log(Level.SEVERE, null, ex);
            }
        } catch (SQLiteException ex) {
            Logger.getLogger(FXMLDocumentController.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Convience method for fetching tableView vertical ScrollBar
     *
     * @param table table to fetch vertical ScrollBar from
     * @return vertical ScrollBar
     */
    private ScrollBar getVerticalScrollbar(TableView<?> table) {
        ScrollBar result = null;
        for (Node n : table.lookupAll(".scroll-bar")) {
            if (n instanceof ScrollBar) {
                ScrollBar bar = (ScrollBar) n;
                if (bar.getOrientation().equals(Orientation.VERTICAL)) {
                    result = bar;
                }
            }
        }
        return result;
    }

    @Override
    public void initialize(URL url, ResourceBundle rb) {
        //DB Files path loading if saved
        prop = new Properties();
        InputStream input = null;
        try {

            input = new FileInputStream("config.properties");
            //load a properties file from class path, inside static method
            prop.load(input);
            dbPath = new File(prop.getProperty("dbPath"));
            nyaaPath = new File(prop.getProperty("nyaaPath"));
        } catch (IOException ex) {
        } finally {
            if (input != null) {
                try {
                    input.close();
                } catch (IOException e) {
                }
            }
        }

        //path choose dialog; TODO on demand ----------
        dbChooserDialog = new Dialog();
        dbChooserDialog.setTitle("Select dump files");
        dbChooserDialog.getDialogPane().setHeader(null);
        dbChooserDialog.setResizable(true);
        Label helpText = new Label("Select the location of the view.sqlite3 and nyaa.txt files. (unzip nyaa.txt.bz2)");
        Label viewLabel = new Label("view.sqlite3: ");
        Label nyaaLabel = new Label("nyaa.txt: ");
        viewField = new TextField();
        nyaaField = new TextField();
        Button viewButton = new Button("...");
        Button nyaaButton = new Button("...");
        viewButton.setOnMousePressed((e) -> {
            dbPath = fileChooser.showOpenDialog(dbChooserDialog.getOwner());
            viewField.setText(dbPath.toString());
        });
        nyaaButton.setOnMousePressed((MouseEvent e) -> {
            nyaaPath = fileChooser.showOpenDialog(dbChooserDialog.getOwner());
            nyaaField.setText(nyaaPath.toString());
        });

        GridPane dbChooserDialogGrid = new GridPane();
        dbChooserDialogGrid.add(helpText, 2, 1);

        dbChooserDialogGrid.add(viewLabel, 1, 2);
        dbChooserDialogGrid.add(viewField, 2, 2);
        dbChooserDialogGrid.add(nyaaLabel, 1, 3);
        dbChooserDialogGrid.add(nyaaField, 2, 3);
        dbChooserDialogGrid.add(viewButton, 3, 2);
        dbChooserDialogGrid.add(nyaaButton, 3, 3);

        dbChooserDialog.getDialogPane().setContent(dbChooserDialogGrid);
        dbChooserDialog.getDialogPane().getButtonTypes().add(ButtonType.OK);
        //----------------------

        //Checks if db paths loaded via properties, loads them via dialog otherwise
        if (nyaaPath == null || dbPath == null) {
            showDbChooserDialog();
        } else {
            setBD();
        }

        copyTooltip.setAutoHide(true);

        //help/about panel
        help = new Alert(Alert.AlertType.INFORMATION, "  Nyanner - Nyaa dump brownser\n"
                + "  Get the latest version at https://yanju.moe/nyanner\n"
                + "  Copyright © 2017 Yanju <yanju@yanju.moe>\n"
                + "  This work is free. You can redistribute it and/or modify it under the "
                + "terms of the Do What The Fuck You Want To Public License, Version 2, "
                + "as published by Sam Hocevar. See http://www.wtfpl.net/ for more details.\n"
                + " \n"
                + "  @version 0.5 04/05/2017");
        help.setHeaderText("About Nyanner");
        help.setTitle("About Nyanner");
        //-------

        //sets focus to search field
        searchField.requestFocus();
        itemsTable.setPlaceholder(new Label("Type and press enter or click search"));

        //hack to unser DB is disposed on close TODO better
        searchField.sceneProperty().addListener((a, b, c) -> {
            c.windowProperty().addListener((d, e, f) -> {
                f.setOnCloseRequest((g) -> {
                    disposeDB();
                });
            });
        });

        //Load category images
        categoryImages = new HashMap<>();
        categoryImages.put("1_32", new Image("/resources/www-32.png"));
        categoryImages.put("1_37", new Image("/resources/www-37.png"));
        categoryImages.put("1_38", new Image("/resources/www-38.png"));
        categoryImages.put("1_11", new Image("/resources/www-11.png"));
        categoryImages.put("3_14", new Image("/resources/www-14.png"));
        categoryImages.put("3_15", new Image("/resources/www-15.png"));
        categoryImages.put("2_12", new Image("/resources/www-12.png"));
        categoryImages.put("2_39", new Image("/resources/www-39.png"));
        categoryImages.put("2_13", new Image("/resources/www-13.png"));
        categoryImages.put("5_19", new Image("/resources/www-19.png"));
        categoryImages.put("5_22", new Image("/resources/www-22.png"));
        categoryImages.put("5_21", new Image("/resources/www-21.png"));
        categoryImages.put("5_20", new Image("/resources/www-20.png"));
        categoryImages.put("4_18", new Image("/resources/www-18.png"));
        categoryImages.put("4_17", new Image("/resources/www-17.png"));
        categoryImages.put("6_23", new Image("/resources/www-23.png"));
        categoryImages.put("6_24", new Image("/resources/www-24.png"));
        //--

        //Load category codes
        categoryCodes = new HashMap<>();
        categoryCodes.put("All categories", "0_0");
        categoryCodes.put("Anime", "nyaa.category like '1_%'");
        categoryCodes.put("Anime - Anime Music Video", "nyaa.category='1_32'");
        categoryCodes.put("Anime - English-translated", "nyaa.category='1_37'");
        categoryCodes.put("Anime - Non-English-translated", "nyaa.category='1_38'");
        categoryCodes.put("Anime - Raw", "nyaa.category='1_11'");
        categoryCodes.put("Audio", "nyaa.category like '3_%'");
        categoryCodes.put("Audio - Lossless", "nyaa.category='3_14'");
        categoryCodes.put("Audio - Lossy", "nyaa.category='3_15'");
        categoryCodes.put("Literature", "nyaa.category like '2_%'");
        categoryCodes.put("Literature - English-translated", "nyaa.category='2_12'");
        categoryCodes.put("Literature - Non-English-translated", "nyaa.category='2_39'");
        categoryCodes.put("Literature - Raw", "nyaa.category='2_13'");
        categoryCodes.put("Live Action", "nyaa.category like '5_%'");
        categoryCodes.put("Live Action - English-translated", "nyaa.category='5_19'");
        categoryCodes.put("Live Action - Idol/Promotional Video", "nyaa.category='5_22'");
        categoryCodes.put("Live Action - Non-English-translated", "nyaa.category='5_21'");
        categoryCodes.put("Live Action - Raw", "nyaa.category='5_20'");
        categoryCodes.put("Pictures", "nyaa.category like '4_%'");
        categoryCodes.put("Pictures - Graphics", "nyaa.category='4_18'");
        categoryCodes.put("Pictures - Photos", "nyaa.category='4_17'");
        categoryCodes.put("Software", "nyaa.category like '6_%'");
        categoryCodes.put("Software - Applications", "nyaa.category='6_23'");
        categoryCodes.put("Software - Games", "nyaa.category='6_24'");
        //--

        //sets table columns and cell factories
        magnetIcon = new Image("/resources/magnetIcon.png");
        torrentList = FXCollections.observableArrayList();
        itemsTable.setItems(torrentList);
        categoryColumn.setCellValueFactory(new PropertyValueFactory<>("category"));
        categoryColumn.setCellFactory(value -> new CategoryCell<>());
        nameColumn.setCellValueFactory(new PropertyValueFactory<>("name"));
        nameColumn.setStyle("-fx-text-fill: #106655; -fx-font-weight: bold; -fx-font-size: 12px;");
        sizeColumn.setCellValueFactory(new PropertyValueFactory<>("size"));
        dlColumn.setCellValueFactory(new PropertyValueFactory<>("id"));

        dlColumn.setCellFactory(value -> {
            MagnetCell<Torrent> cell = new MagnetCell<>(magnetIcon);
            cell.setOnMouseEntered((e) -> {
                if (cell.getItem() != null) {
                    cell.setCursor(Cursor.HAND);
                }
            });
            cell.setOnMouseClicked((e) -> {
                if (cell.getItem() != null) {
                    download(cell.getItem());
                    copyTooltip.show(cell.getScene().getWindow());
                }

            });
            return cell;
        });
        itemsTable.setRowFactory(e
                -> new TableRow<Torrent>() {
            @Override
            public void updateItem(Torrent item, boolean empty) {
                super.updateItem(item, empty);
                if (item == null) {
                    setStyle("");
                } else if (item.getQuality().equals("trusted")) {
                    setStyle("-fx-background-color: #98d9a8;");
                } else if (item.getQuality().equals("remake")) {
                    setStyle("-fx-background-color: #f0b080;");
                } else if (item.getQuality().equals("aplus")) {
                    setStyle("-fx-background-color: #60b0f0;");
                } else {
                    setStyle("-fx-background-color: #FFFFFF;");
                }
            }

        });

        //searchfield enter key listener
        searchField.setOnKeyPressed((event) -> {
            if (event.getCode() == KeyCode.ENTER) {
                search();
            }
        });

    }

}
